from django.db import models
from pyvalem.formula import Formula
from pyvalem.stateful_species import StatefulSpecies

from _utils.models import QualifiedIDMixin


class Species(QualifiedIDMixin, models.Model):
    qid_prefix = "F"

    id = models.AutoField(primary_key=True)

    text = models.CharField(max_length=80)
    html = models.CharField(max_length=200)
    charge = models.SmallIntegerField(default=0, null=True)

    class Meta:
        verbose_name_plural = "Species"

    def __str__(self):
        return self.text

    @classmethod
    def get_from_text(cls, text):
        """Looks for a Species with equivalent canonicalised version of the
        text. Uses pyvalem Formula.__repr__ for the canonicalisation.
        If not present, Species.DoesNotExist is raised.

        Parameters
        ----------
        text : str

        Returns
        -------
        Species
        """
        text_can = repr(Formula(text))
        return cls.objects.get(text=text_can)

    @classmethod
    def get_or_create_from_text(cls, text):
        """Looks for a Species with equivalent canonicalised version of the
        text.

        Parameters
        ----------
        text : str

        Returns
        -------
        (Species, bool)
        """
        pyvalem_formula = Formula(text)
        text_can = repr(pyvalem_formula)
        try:
            return cls.objects.get(text=text_can), False
        except cls.DoesNotExist:
            # re-instantiate the pyvalem_formula with canonicalised
            # text to canonicalise html also:
            pyvalem_formula = Formula(text_can)
            species = cls.objects.create(
                text=text_can, charge=pyvalem_formula.charge, html=pyvalem_formula.html
            )
            return species, True


class RP(QualifiedIDMixin, models.Model):
    qid_prefix = "RP"

    id = models.AutoField(primary_key=True)
    species = models.ForeignKey(Species, on_delete=models.CASCADE)

    text = models.CharField(max_length=200)
    html = models.CharField(max_length=600)

    def __str__(self):
        return self.text

    @classmethod
    def get_from_text(cls, text):
        """Looks for RP with equivalent canonicalised version of the
        text. Uses pyvalem StatefulSpecies.__repr__ for the canonicalisation.
        If not present, RP.DoesNotExist is raised.

        Parameters
        ----------
        text : str

        Returns
        -------
        RP
        """
        text_can = repr(StatefulSpecies(text))
        return cls.objects.get(text=text_can)

    @classmethod
    def get_or_create_from_text(cls, text):
        """Looks for a Species with equivalent canonicalised version of the
        text.

        Parameters
        ----------
        text : str

        Returns
        -------
        (RP, bool)
        """
        pyvalem_stateful_species = StatefulSpecies(text)
        text_can = repr(pyvalem_stateful_species)
        try:
            return cls.objects.get(text=text_can), False
        except cls.DoesNotExist:
            pyvalem_formula = pyvalem_stateful_species.formula
            species, _ = Species.get_or_create_from_text(repr(pyvalem_formula))
            # re-instantiate the pyvalem_stateful_species with canonicalised
            # text to canonicalise html also and sort the states consistently
            # with the text and html:
            pyvalem_stateful_species = StatefulSpecies(text_can)
            # build the RP instance:
            rp = cls.objects.create(
                species=species, text=text_can, html=pyvalem_stateful_species.html
            )
            # attach the states:
            for pyvalem_state in pyvalem_stateful_species.states:
                State.objects.create(
                    rp=rp,
                    text=repr(pyvalem_state),
                    html=pyvalem_state.html,
                    state_type=State.STATE_TYPE_MAP[pyvalem_state.__class__.__name__],
                )
            return rp, True

    @property
    def charge(self):
        """Returns the charge of RPs Species."""
        return self.species.charge


class State(QualifiedIDMixin, models.Model):
    qid_prefix = "S"

    KEY_VALUE_PAIR = 0
    GENERIC_EXCITED_STATE = 1
    ATOMIC_CONFIGURATION = 2
    ATOMIC_TERM_SYMBOL = 3
    DIATOMIC_MOLECULAR_CONFIGURATION = 4
    MOLECULAR_TERM_SYMBOL = 5
    VIBRATIONAL_STATE = 6
    ROTATIONAL_STATE = 10
    RACAH_SYMBOL = 11

    STATE_TYPE_CHOICES = (
        (KEY_VALUE_PAIR, "KeyValuePair"),
        (GENERIC_EXCITED_STATE, "GenericExcitedState"),
        (ATOMIC_CONFIGURATION, "AtomicConfiguration"),
        (ATOMIC_TERM_SYMBOL, "AtomicTermSymbol"),
        (DIATOMIC_MOLECULAR_CONFIGURATION, "DiatomicMolecularConfiguration"),
        (MOLECULAR_TERM_SYMBOL, "MolecularTermSymbol"),
        (VIBRATIONAL_STATE, "VibrationalState"),
        (ROTATIONAL_STATE, "RotationalState"),
        (RACAH_SYMBOL, "RacahSymbol"),
    )

    STATE_TYPE_MAP = {v: k for k, v in STATE_TYPE_CHOICES}

    id = models.AutoField(primary_key=True)
    rp = models.ForeignKey(RP, on_delete=models.CASCADE)

    state_type = models.SmallIntegerField(choices=STATE_TYPE_CHOICES)
    text = models.CharField(max_length=64)
    html = models.CharField(max_length=100)

    def __str__(self):
        return self.text



# Generated by Django 3.0 on 2022-01-25 16:02

import _utils.models
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="RP",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("text", models.CharField(max_length=200)),
                ("html", models.CharField(max_length=600)),
            ],
            bases=(_utils.models.QualifiedIDMixin, models.Model),
        ),
        migrations.CreateModel(
            name="Species",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("text", models.CharField(max_length=80)),
                ("html", models.CharField(max_length=200)),
                ("charge", models.SmallIntegerField(default=0, null=True)),
            ],
            bases=(_utils.models.QualifiedIDMixin, models.Model),
        ),
        migrations.CreateModel(
            name="State",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "state_type",
                    models.SmallIntegerField(
                        choices=[
                            (0, "KeyValuePair"),
                            (1, "GenericExcitedState"),
                            (2, "AtomicConfiguration"),
                            (3, "AtomicTermSymbol"),
                            (4, "DiatomicMolecularConfiguration"),
                            (5, "MolecularTermSymbol"),
                            (6, "VibrationalState"),
                            (10, "RotationalState"),
                            (11, "RacahSymbol"),
                        ]
                    ),
                ),
                ("text", models.CharField(max_length=64)),
                ("html", models.CharField(max_length=100)),
                (
                    "rp",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="rp.RP"
                    ),
                ),
            ],
            bases=(_utils.models.QualifiedIDMixin, models.Model),
        ),
        migrations.AddField(
            model_name="rp",
            name="species",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="rp.Species"
            ),
        ),
    ]



from django.db import models
from pyvalem.reaction import Reaction as PVReaction
from pyvalem.reaction import ReactionParseError

from _utils.models import QualifiedIDMixin
from rp.models import RP


class ProcessType(QualifiedIDMixin, models.Model):
    """A model class defining the description of a process.

    The processes described by this app fall into categories which are
    described by this model. Each category has a three-letter abbreviation
    code, description and example.

    :abbreviation: abbreviation for the process
    :description: brief text description of the process
    :example_html: a simple, generic example marked up in HTML

    """

    qid_prefix = "P"

    abbreviation = models.CharField(max_length=3, unique=True)
    description = models.CharField(max_length=200)
    example_html = models.CharField(max_length=200, null=True, blank=True)

    def __str__(self):
        return f"{self.abbreviation}"


class Reaction(QualifiedIDMixin, models.Model):
    qid_prefix = "R"

    id = models.AutoField(primary_key=True)

    reactants = models.ManyToManyField(
        RP, related_name="reactants_related", through="ReactantList"
    )
    products = models.ManyToManyField(
        RP, related_name="products_related", through="ProductList"
    )
    process_types = models.ManyToManyField(ProcessType)

    text = models.CharField(max_length=256, editable=False)
    html = models.CharField(max_length=1024, editable=False)
    latex = models.CharField(max_length=1024, editable=False)
    comment = models.CharField(max_length=1024, blank=True)

    def __str__(self):
        return self.text

    @classmethod
    def all_from_text(cls, text, strict=True):
        """Uses pyvalem to get a canonicalised version of the text and filters
        the database objects by that text. If no reactions equivalent to passed
        text are found, returns an empty query.

        Parameters
        ----------
        text : str

        Returns
        -------
        Query
        """
        text_can = repr(PVReaction(text, strict=strict))
        return cls.objects.filter(text=text_can)

    @classmethod
    def get_from_text(
        cls, text, comment="", process_type_abbreviations=(), strict=True
    ):
        """Looks for an RP with equivalent canonicalised version of the text AND
        identical comment AND the same process_type_abbreviations.
        So this should be strictly named "get_from_data" but we'll keep it like this
        for consistency reasons.

        Parameters
        ----------
        text : str
        comment : str
        process_type_abbreviations : tuple of str

        Returns
        -------
        Reaction
        """
        text_can = repr(PVReaction(text, strict=strict))
        all_with_text_and_comment = cls.objects.filter(text=text_can, comment=comment)
        for reaction in all_with_text_and_comment:
            if sorted(pt.abbreviation for pt in reaction.process_types.all()) == sorted(
                process_type_abbreviations
            ):
                return reaction
        raise cls.DoesNotExist

    @classmethod
    def get_or_create_from_text(
        cls, text, comment="", process_type_abbreviations=(), strict=True
    ):
        """

        Parameters
        ----------
        text : str
        comment : str
        process_type_abbreviations : tuple of str

        Returns
        -------
        (Reaction, bool)
        """
        try:
            return (
                cls.get_from_text(text, comment, process_type_abbreviations, strict),
                False,
            )
        except cls.DoesNotExist:
            # canonicalised text and html:
            text_can = repr(PVReaction(text, strict=strict))
            pyvalem_reaction = PVReaction(
                text_can, strict=strict
            )  # to reset the html to canonic.
            html = pyvalem_reaction.html
            latex = pyvalem_reaction.latex
            # create the Reaction object:
            reaction = cls.objects.create(
                text=text_can, html=html, latex=latex, comment=comment
            )
            # populate the reactants and products with RP instances:
            for attr, Intermediate in zip(
                ["reactants", "products"], [ReactantList, ProductList]
            ):
                for stoich, stateful_species in getattr(pyvalem_reaction, attr):
                    for _ in range(stoich):
                        Intermediate.objects.create(
                            reaction=reaction,
                            rp=RP.get_or_create_from_text(repr(stateful_species))[0],
                        )
            # assign the ProcessTypes:
            process_types = [
                ProcessType.objects.get(abbreviation=abbrev)
                for abbrev in process_type_abbreviations
            ]
            for process_type in process_types:
                reaction.process_types.add(process_type)

            return reaction, True

    @property
    def molecularity(self):
        """Return the molecularity of the reaction (number of reactants)."""
        return self.reactants.count()

    def _reset_html(self):
        try:
            pyvalem_reaction = PVReaction(self.text)
        except ReactionParseError:
            pyvalem_reaction = PVReaction(self.text, strict=False)

        self.html = pyvalem_reaction.html
        self.save()


class ReactantList(models.Model):
    """ReactantList implements the ManyToMany relationship between an reaction
    and its reactant RP objects."""

    reaction = models.ForeignKey(Reaction, on_delete=models.CASCADE)
    rp = models.ForeignKey(RP, on_delete=models.CASCADE)

    class Meta:
        db_table = "rxn_reaction_reactants"


class ProductList(models.Model):
    """ProductList implements the ManyToMany relationship between an reaction
    and its product RP objects."""

    reaction = models.ForeignKey(Reaction, on_delete=models.CASCADE)
    rp = models.ForeignKey(RP, on_delete=models.CASCADE)

    class Meta:
        db_table = "rxn_reaction_products"



# Generated by Django 3.0 on 2022-03-14 16:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("rxn", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="reaction",
            name="latex",
            field=models.CharField(default="TODO", editable=False, max_length=1024),
            preserve_default=False,
        ),
    ]



from django.db import models
from refs.models import Ref

from _utils.models import ProvenanceMixin, QualifiedIDMixin
from rxn.models import Reaction


class ReactionDataSet(QualifiedIDMixin, ProvenanceMixin, models.Model):
    """An Abstract Base Class representing a dataset.

    Links a `Reaction` instance and one or more `Ref` reference instances.
    The subclasses are responsible for implementing any data fields relevant to their
    scope.
    """

    qid_prefix = "D"

    id = models.AutoField(primary_key=True)
    reaction = models.ForeignKey(Reaction, on_delete=models.CASCADE)
    refs = models.ManyToManyField(Ref)

    class Meta:
        abstract = True

    def __str__(self):
        return str(self.reaction)



# Generated by Django 3.0 on 2022-01-25 16:02

import _utils.models
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("rp", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="ProcessType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("abbreviation", models.CharField(max_length=3, unique=True)),
                ("description", models.CharField(max_length=200)),
                (
                    "example_html",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
            ],
            bases=(_utils.models.QualifiedIDMixin, models.Model),
        ),
        migrations.CreateModel(
            name="ProductList",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "db_table": "rxn_reaction_products",
            },
        ),
        migrations.CreateModel(
            name="ReactantList",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "db_table": "rxn_reaction_reactants",
            },
        ),
        migrations.CreateModel(
            name="Reaction",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("text", models.CharField(editable=False, max_length=256)),
                ("html", models.CharField(editable=False, max_length=1024)),
                ("comment", models.CharField(blank=True, max_length=1024)),
                ("process_types", models.ManyToManyField(to="rxn.ProcessType")),
                (
                    "products",
                    models.ManyToManyField(
                        related_name="products_related",
                        through="rxn.ProductList",
                        to="rp.RP",
                    ),
                ),
                (
                    "reactants",
                    models.ManyToManyField(
                        related_name="reactants_related",
                        through="rxn.ReactantList",
                        to="rp.RP",
                    ),
                ),
            ],
            bases=(_utils.models.QualifiedIDMixin, models.Model),
        ),
        migrations.AddField(
            model_name="reactantlist",
            name="reaction",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="rxn.Reaction"
            ),
        ),
        migrations.AddField(
            model_name="reactantlist",
            name="rp",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="rp.RP"
            ),
        ),
        migrations.AddField(
            model_name="productlist",
            name="reaction",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="rxn.Reaction"
            ),
        ),
        migrations.AddField(
            model_name="productlist",
            name="rp",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="rp.RP"
            ),
        ),
    ]



from setuptools import setup, find_packages
from pathlib import Path

root = Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (root / "README.rst").read_text(encoding="utf-8")

setup(
    name="django-valem",
    version="0.1.7",
    description="A collection of Django apps defining data models for managing "
    "chemical species, reactions and datasets.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/xnx/django-valem",
    author="Christian Hill",
    author_email="ch.hill@iaea.org",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Framework :: Django",
        "Operating System :: OS Independent",
    ],
    keywords="django, chemistry, formula, species, state, reaction",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        "pyvalem>=2.5.9",
        "django-pyref>=0.5.1",
    ],
    extras_require={"dev": ["black", "coverage", "django==3", "ipython"]},
    project_urls={
        "Bug Reports": "https://github.com/xnx/django-valem/issues",
    },
)



from django.db import models
from django.db import models

# noinspection PyUnresolvedReferences
class QualifiedIDMixin:
    qid_prefix = "-"

    @property
    def qualified_id(self):
        return f"{self.qid_prefix}{self.id}"

    def __repr__(self):
        """By default, prepend the qualified ID to the class' __repr__()"""
        return f"<{self.qualified_id}: {self}>"


class ProvenanceMixin(models.Model):
    added_by_user_id = models.IntegerField(null=True, blank=True)
    time_added = models.DateTimeField(auto_now_add=True)
    time_modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True
